using System;
using System.IO;
using System.Linq;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.Threading.Tasks;
using NUnit.Framework;

using Smdn.Net.Pop3.Protocol;
using Smdn.Net.Pop3.Protocol.Client;

namespace Smdn.Net.Pop3.Client.Session {
  [TestFixture]
  public class PopSessionTests {
    [Test]
    public void TestConnect()
    {
      PopPseudoServer.Start((server) => {
        using (var session = new PopSession(server.Host, server.Port)) {
          Assert.IsFalse(session.ApopAvailable);
          Assert.IsTrue(string.IsNullOrEmpty(session.Timestamp));
          Assert.AreEqual(PopSessionState.Authorization, session.State);
          Assert.AreEqual(string.Format("pop://{0}:{1}/", server.Host, server.Port),
                          session.Authority.AbsoluteUri);
          Assert.AreEqual(0, session.ServerCapabilities.Count);
        }
      });
    }

    [Test]
    public void TestConnectTransactionTimeout()
    {
      PopPseudoServer.StartNotAccepting((server) => {
        server.SetGreetingResponse(null); // no greeting response

        var ex = Assert.Throws<PopConnectionException>(() => {
          using (var session = new PopSession(server.Host, server.Port, 500)) {
          }
        });

        Assert.IsInstanceOf<TimeoutException>(ex.InnerException);
      });
    }

    [Test]
    public void TestConnectErr()
    {
      PopPseudoServer.Start((server) => {
        server.SetGreetingResponse("-ERR try again\r\n", null, null, true);

        var ex = Assert.Throws<PopConnectionException>(() => {
          using (var session = new PopSession(server.Host, server.Port)) {
          }
        });

        Assert.IsNull(ex.InnerException);
        StringAssert.Contains("try again", ex.Message);
      });
    }

    [Test]
    public void TestConnectInvalidResponse()
    {
      PopPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK pseudo IMAP server ready\r\n");

        var ex = Assert.Throws<PopConnectionException>(() => {
          using (var session = new PopSession(server.Host, server.Port)) {
          }
        });

        Assert.IsNull(ex.InnerException);
        StringAssert.Contains("unexpected response", ex.Message);
      });
    }

    [Test]
    public void TestConnectConnectionTimeout()
    {
      var endPoint = new IPEndPoint(IPAddress.Loopback, 0);

      using (var server = new Socket(endPoint.AddressFamily, SocketType.Stream, ProtocolType.Tcp)) {
        server.Bind(endPoint);
        server.Listen(1);

        var ex = Assert.Throws<PopConnectionException>(() => {
          var serverEndPoint = server.LocalEndPoint as IPEndPoint;

          using (var session = new PopSession(serverEndPoint.Address.ToString(), serverEndPoint.Port, 1000)) {
          }
        });

#if NETFRAMEWORK || NETSTANDARD2_0
        server.Close();
#else
        server.Dispose();
#endif

        Assert.IsInstanceOf<TimeoutException>(ex.InnerException);
      }
    }

    [Test]
    public void TestConnectConnectionTimeout_ServerNotAccepting()
    {
      PopPseudoServer.StartNotAccepting((server) => {
        var ex = Assert.Throws<PopConnectionException>(() => {
          using (var session = new PopSession(server.Host, server.Port, 500)) {
          }
        });

        Assert.IsInstanceOf<TimeoutException>(ex.InnerException);
      });
    }

    [Test]
    public void TestConnectSocketError()
    {
      var ex = Assert.Throws<PopConnectionException>(() => {
        using (var session = new PopSession("pop.invalid", 10110)) {
        }
      });

      Assert.IsInstanceOf<System.Net.Sockets.SocketException>(ex.InnerException);
    }

    [Test]
    public void TestConnectSsl()
    {
      PopPseudoServer.Start((server) => {
        var streamUpgraded = false;

        using (var session = new PopSession(server.Host,
                                            server.Port,
                                            createAuthenticatedStreamCallback: (connection, baseStream) => { streamUpgraded = true; return baseStream; /* TODO: return SslStream */ })) {
          Assert.IsFalse(session.ApopAvailable);
          Assert.IsTrue(string.IsNullOrEmpty(session.Timestamp));
          Assert.AreEqual(PopSessionState.Authorization, session.State);
          Assert.AreEqual(string.Format("pops://{0}:{1}/", server.Host, server.Port),
                          session.Authority.AbsoluteUri);
          Assert.AreEqual(0, session.ServerCapabilities.Count);
          Assert.IsTrue(streamUpgraded, "stream upgraded");
          Assert.IsTrue(session.IsSecureConnection);
        }
      });
    }

    [Test]
    public void TestConnectSslExceptionWhileUpgrading()
    {
      PopPseudoServer.Start((server) => {
        try {
          using (var session = new PopSession(server.Host,
                                              server.Port,
                                              createAuthenticatedStreamCallback: (connection, baseStream) => throw new System.Security.Authentication.AuthenticationException())) {
            Assert.Fail("connected");
          }
        }
        catch (PopUpgradeConnectionException ex) {
          Assert.IsNotNull(ex.InnerException);
          Assert.IsInstanceOf<System.Security.Authentication.AuthenticationException>(ex.InnerException);
        }
      });
    }

    [Test]
    public void TestDisconnect()
    {
      PopPseudoServer.Start((server) => {
        var session = new PopSession(server.Host, server.Port);

        session.Disconnect(false);

        Assert.IsTrue(session.IsDisposed);
        Assert.IsFalse(session.IsTransactionProceeding);

        Assert.Throws<ObjectDisposedException>(() => session.NoOp());

        Assert.Throws<ObjectDisposedException>(() => { Assert.Fail("State = {0}", session.State); });
        Assert.Throws<ObjectDisposedException>(() => { Assert.Fail("SendTimeout = {0}", session.SendTimeout); });
        Assert.Throws<ObjectDisposedException>(() => { Assert.Fail("ReceiveTimeout = {0}", session.ReceiveTimeout); });
        Assert.Throws<ObjectDisposedException>(() => { Assert.Fail("TransactionTimeout = {0}", session.TransactionTimeout); });
      });
    }

    [Test]
    public void TestDisconnectedFromServer()
    {
      PopPseudoServer.Start((server) => {
        server.SetGreetingResponse("+OKr\n", null, null, true);

        using (var session = new PopSession(server.Host, server.Port)) {
          Assert.Throws<PopConnectionException>(() => session.Quit());

          Assert.AreEqual(PopSessionState.NotConnected, session.State);
          Assert.AreEqual(null, session.Authority);

          Assert.Throws<PopProtocolViolationException>(() => session.NoOp());
        }
      });
    }

    [Test]
    public void TestGreetingWithTimestamp()
    {
      PopPseudoServer.Start((server) => {
        server.SetGreetingResponse("+OK POP3 server ready <1896.697170952@dbc.mtview.ca.us>\r\n");

        using (var session = new PopSession(server.Host, server.Port)) {
          Assert.IsTrue(session.ApopAvailable);
          Assert.AreEqual("<1896.697170952@dbc.mtview.ca.us>", session.Timestamp);
          Assert.AreEqual(PopSessionState.Authorization, session.State);
          Assert.AreEqual(string.Format("pop://{0}:{1}/", server.Host, server.Port),
                          session.Authority.AbsoluteUri);
          Assert.AreEqual(0, session.ServerCapabilities.Count);
        }
      });
    }

    [Test]
    public void TestGreetingWithTimestamp_RespondOnlyTimestamp()
    {
      PopPseudoServer.Start((server) => {
        server.SetGreetingResponse("+OK <1896.697170952@dbc.mtview.ca.us>\r\n");

        using (var session = new PopSession(server.Host, server.Port)) {
          Assert.IsTrue(session.ApopAvailable);
          Assert.AreEqual("<1896.697170952@dbc.mtview.ca.us>", session.Timestamp);
          Assert.AreEqual(PopSessionState.Authorization, session.State);
          Assert.AreEqual(string.Format("pop://{0}:{1}/", server.Host, server.Port),
                          session.Authority.AbsoluteUri);
          Assert.AreEqual(0, session.ServerCapabilities.Count);
        }
      });
    }

    [Test]
    public void TestChangeReadOnlyPropertyServerCapabilities()
    {
      TestCases.WithConnectedSession(session => {
        Assert.Throws<NotSupportedException>(() => session.ServerCapabilities.Add(PopCapability.User));
      });
    }

    [Test]
    public void TestSetReceiveTimeoutZero()
    {
      TestCases.WithConnectedSession(session => {
        session.ReceiveTimeout = 0;

        Assert.AreEqual(0, session.ReceiveTimeout);
      });
    }

    [Test]
    public void TestSetReceiveTimeoutInfinite()
    {
      TestCases.WithConnectedSession(session => {
        session.ReceiveTimeout = Timeout.Infinite;

        Assert.AreEqual(Timeout.Infinite, session.ReceiveTimeout);
      });
    }

    [Test]
    public void TestSetSendTimeoutZero()
    {
      TestCases.WithConnectedSession(session => {
        session.SendTimeout = 0;

        Assert.AreEqual(0, session.SendTimeout);
      });
    }

    [Test]
    public void TestSetSendTimeoutInfinite()
    {
      TestCases.WithConnectedSession(session => {
        session.SendTimeout = Timeout.Infinite;

        Assert.AreEqual(Timeout.Infinite, session.SendTimeout);
      });
    }

    [Test]
    public void TestSetTransactionTimeoutZero()
    {
      TestCases.WithConnectedSession(session => {
        session.TransactionTimeout = 0;

        Assert.AreEqual(0, session.TransactionTimeout);
      });
    }

    [Test]
    public void TestSetTransactionTimeoutInfinite()
    {
      TestCases.WithConnectedSession(session => {
        session.TransactionTimeout = Timeout.Infinite;

        Assert.AreEqual(Timeout.Infinite, session.TransactionTimeout);
      });
    }

    [Test]
    public void TestSetTransactionTimeoutWhileTransactionProceeding()
    {
      TestCases.WithConnectedSession((session, server) => {
        session.TransactionTimeout = 250;
        session.SendTimeout = System.Threading.Timeout.Infinite;
        session.ReceiveTimeout = System.Threading.Timeout.Infinite;

        // not respond CAPA response termination
        server.EnqueueResponse("CAPA\r\n",
                               "+OK List of capabilities follows\r\n");

        Assert.Throws<TimeoutException>(() => session.Capa());

        Assert.AreEqual(PopSessionState.NotConnected, session.State);

        Assert.Throws<PopProtocolViolationException>(() => session.NoOp());
      });
    }

    [Test]
    public void TestSetSocketTimeoutWhileTransactionProceeding()
    {
      TestCases.WithConnectedSession((session, server) => {
        session.TransactionTimeout = System.Threading.Timeout.Infinite;
        session.SendTimeout = 250;
        session.ReceiveTimeout = 250;

        // not respond CAPA
        //server.EnqueueResponse("+OK List of capabilities follows\r\n.\r\n");

        Assert.Throws<TimeoutException>(() => session.Capa());

        Assert.AreEqual(PopSessionState.NotConnected, session.State);

        Assert.Throws<PopProtocolViolationException>(() => session.NoOp());
      });
    }

    [Test]
    public void TestInternalErrorWhileTransactionProceeding()
    {
      TestCases.WithConnectedSession((session, server) => {
        server.EnqueueResponse("CAPA\r\n",
                               "+OK List of capabilities follows\r\n" +
                               "\r\n" + // invalid response
                               ".\r\n");

        var ex = Assert.Throws<PopException>(() => session.Capa());

        Assert.IsNotNull(ex.InnerException);

        Assert.AreEqual(PopSessionState.NotConnected, session.State);

        Assert.Throws<PopProtocolViolationException>(() => session.NoOp());
      });
    }

    [Test]
    public void TestBeginNewTransactionWhileTransactionProceeding()
    {
      TestCases.WithConnectedSession((session, server) => {
        using (var transactionStartedEvent = new ManualResetEvent(false)) {
          server.EnqueueResponse("CAPA\r\n",
                                 "+OK Capability list follows\r\n", // incomplete response
                                 null,
                                 () => transactionStartedEvent.Set());

          var capaTask = Task.Run(() => session.Capa());

          transactionStartedEvent.WaitOne();

          Assert.IsTrue(session.IsTransactionProceeding);

          Assert.Throws<InvalidOperationException>(() => session.Capa());

          server.EnqueuePushResponse(".\r\n");

          Assert.DoesNotThrow(() => {
            if (!capaTask.Wait(TimeSpan.FromSeconds(1)))
              Assert.Fail("transaction not stopped");
          });
        }
      });
    }

    [Test]
    public void TestQuitInAuthorizationState()
    {
      TestCases.WithConnectedSession((session, server) => {
        server.EnqueueResponse("QUIT\r\n",
                               "+OK dewey POP3 server signing off\r\n",
                               true);

        Assert.IsTrue((bool)session.Quit());

        Assert.AreEqual(PopSessionState.NotConnected, session.State);
        Assert.IsNull(session.Authority);
      });
    }

    [Test]
    public void TestQuitAfterDisconnected()
    {
      TestCases.WithConnectedSession((session, server) => {
        server.EnqueueResponse("QUIT\r\n",
                               "+OK dewey POP3 server signing off\r\n",
                               true);

        Assert.IsTrue((bool)session.Quit());

        Assert.AreEqual(PopSessionState.NotConnected, session.State);

        Assert.IsTrue((bool)session.Quit());

        Assert.AreEqual(PopSessionState.NotConnected, session.State);
      });
    }

    [Test]
    public void TestCapa()
    {
      TestCases.WithConnectedSession((session, server) => {
        Assert.IsNotNull(session.ServerCapabilities);
        Assert.AreEqual(0, session.ServerCapabilities.Count);

        server.EnqueueResponse("CAPA\r\n",
                               "+OK Capability list follows\r\n" +
                               "TOP\r\n" +
                               "USER\r\n" +
                               "SASL CRAM-MD5 KERBEROS_V4\r\n" +
                               "RESP-CODES\r\n" +
                               "LOGIN-DELAY 900\r\n" +
                               "PIPELINING\r\n" +
                               "EXPIRE 60\r\n" +
                               "UIDL\r\n" +
                               "IMPLEMENTATION Shlemazle-Plotz-v302\r\n" +
                               ".\r\n");

        PopCapabilitySet capabilities;

        Assert.IsTrue((bool)session.Capa(out capabilities));

        Assert.AreEqual(9, capabilities.Count);
        Assert.IsTrue(capabilities.Contains(PopCapability.Top));
        Assert.IsTrue(capabilities.Contains(PopCapability.User));
        Assert.IsTrue(capabilities.Contains(new PopCapability("SASL", "CRAM-MD5", "KERBEROS_V4")));
        Assert.IsTrue(capabilities.Contains(PopCapability.RespCodes));
        Assert.IsTrue(capabilities.Contains(new PopCapability((string)PopCapability.LoginDelay, "900")));
        Assert.IsTrue(capabilities.Contains(PopCapability.Pipelining));
        Assert.IsTrue(capabilities.Contains(new PopCapability((string)PopCapability.Expire, "60")));
        Assert.IsTrue(capabilities.Contains(PopCapability.Uidl));
        Assert.IsTrue(capabilities.Contains(new PopCapability((string)PopCapability.Implementation, "Shlemazle-Plotz-v302")));

        Assert.IsTrue(capabilities.IsCapable(PopCapability.Sasl));
        Assert.IsTrue(capabilities.IsCapable(PopCapability.LoginDelay));
        Assert.IsTrue(capabilities.IsCapable(PopCapability.Expire));
        Assert.IsTrue(capabilities.IsCapable(PopCapability.Implementation));
        Assert.IsTrue(capabilities.IsCapable(PopAuthenticationMechanism.CRAMMD5));
        Assert.IsTrue(capabilities.IsCapable(PopAuthenticationMechanism.KerberosV4));

        Assert.IsNotNull(session.ServerCapabilities);
        Assert.AreEqual(9, session.ServerCapabilities.Count);
        Assert.IsTrue(session.ServerCapabilities.Contains(PopCapability.User));
        Assert.IsTrue(session.ServerCapabilities.IsCapable(PopCapability.Sasl));

        Assert.Throws<NotSupportedException>(() => session.ServerCapabilities.Add(PopCapability.Lang));
      });
    }

    [Test]
    public void TestQuitInTransactionState()
    {
      TestCases.WithLoggedInSession((session, server) => {
        server.EnqueueResponse("QUIT\r\n",
                               "+OK dewey POP3 server signing off (maildrop empty)\r\n",
                               true);

        Assert.IsTrue((bool)session.Quit());

        Assert.AreEqual(PopSessionState.NotConnected, session.State);
        Assert.IsNull(session.Authority);
      });
    }

    [Test]
    public void TestGenericCommandWithNoArguments()
    {
      TestCases.WithConnectedSession((session, server) => {
        // generic command transaction
        server.EnqueueResponse("X-EXT-COMMAND\r\n",
                               "+OK done\r\n");

        PopResponse[] responses;

        Assert.IsTrue((bool)session.GenericCommand("X-EXT-COMMAND", out responses));

        Assert.AreEqual(1, responses.Length);
        Assert.IsInstanceOf<PopStatusResponse>(responses[0]);

        var status = responses[0] as PopStatusResponse;

        Assert.AreEqual("done", status.Text);
      });
    }

    [Test]
    public void TestGenericCommandWithArguments()
    {
      TestCases.WithConnectedSession((session, server) => {
        // generic command transaction
        server.EnqueueResponse("X-EXT-COMMAND arg1 arg2 arg3\r\n",
                               "+OK done\r\n");

        PopResponse[] responses;

        Assert.IsTrue((bool)session.GenericCommand("X-EXT-COMMAND", out responses, "arg1", "arg2", "arg3"));

        Assert.AreEqual(1, responses.Length);
        Assert.IsInstanceOf<PopStatusResponse>(responses[0]);

        var status = responses[0] as PopStatusResponse;

        Assert.AreEqual("done", status.Text);
      });
    }

    [Test]
    public void TestGenericCommandMultilineResponse()
    {
      TestCases.WithConnectedSession((session, server) => {
        // generic command transaction
        server.EnqueueResponse("X-EXT-COMMAND\r\n",
                               "+OK [X-RESP-CODE] done\r\n" +
                               "1st line\r\n" +
                               "2nd line\r\n" +
                               ".byte-stuffed line\r\n" +
                               ".\r\n");

        PopResponse[] responses;

        Assert.IsTrue((bool)session.GenericCommand("X-EXT-COMMAND", true, out responses));

        Assert.AreEqual(5, responses.Length);
        Assert.IsInstanceOf<PopStatusResponse>(responses[0]);
        Assert.IsInstanceOf<PopFollowingResponse>(responses[1]);
        Assert.IsInstanceOf<PopFollowingResponse>(responses[2]);
        Assert.IsInstanceOf<PopFollowingResponse>(responses[3]);
        Assert.IsInstanceOf<PopTerminationResponse>(responses[4]);

        var status = responses[0] as PopStatusResponse;

        Assert.AreEqual("done", status.Text);
        Assert.IsNotNull(status.ResponseText);
        Assert.AreEqual("X-RESP-CODE", (string)status.ResponseText.Code);

        Assert.AreEqual("1st line", (responses[1] as PopFollowingResponse).Text.ToString());
        Assert.AreEqual("2nd line", (responses[2] as PopFollowingResponse).Text.ToString());
        Assert.AreEqual("byte-stuffed line", (responses[3] as PopFollowingResponse).Text.ToString());
      });
    }

    [Test]
    public void TestSerializeBinaryCommandResultCodeOk()
    {
      TestCases.WithConnectedSession((session, server) => {
        server.EnqueueResponse("CAPA\r\n",
                               "+OK Capability list follows\r\n" +
                               "TOP\r\n" +
                               ".\r\n");

        var result = session.Capa();

        TestUtils.Assert.IsSerializableBinaryFormat(result, delegate(PopCommandResult deserialized) {
          Assert.AreEqual(PopCommandResultCode.Ok, deserialized.Code);
          Assert.IsNotNull(deserialized.StatusResponse);
          Assert.AreEqual(PopStatusIndicator.Positive, deserialized.StatusResponse.Status);
          Assert.AreEqual("Capability list follows", deserialized.StatusResponse.Text);
          Assert.IsNull(deserialized.Description);

          var responses = deserialized.ReceivedResponses.ToArray();

          Assert.AreEqual(3, responses.Length);
          Assert.IsInstanceOf<PopStatusResponse>(responses[0]);
          Assert.IsInstanceOf<PopFollowingResponse>(responses[1]);
          Assert.IsInstanceOf<PopTerminationResponse>(responses[2]);
        });
      });
    }

    [Test]
    public void TestSerializeBinaryCommandResultCodeRequestDone()
    {
      TestCases.WithLoggedInSession((session, server) => {
        var result = session.Login("user", "pass"); // already authenticated

        TestUtils.Assert.IsSerializableBinaryFormat(result, delegate(PopCommandResult deserialized) {
          Assert.AreEqual(PopCommandResultCode.RequestDone, deserialized.Code);
          Assert.IsNull(deserialized.StatusResponse);
          Assert.IsNull(deserialized.ResponseText);
          Assert.IsNotNull(deserialized.Description);

          var responses = deserialized.ReceivedResponses.ToArray();

          Assert.AreEqual(0, responses.Length);
        });
      });
    }

    [Test]
    public void TestSerializeBinaryCommandResultCodeRequestError()
    {
      TestCases.WithConnectedSession(session => {
        var result = session.Login(new NullCredential());

        TestUtils.Assert.IsSerializableBinaryFormat(result, delegate(PopCommandResult deserialized) {
          Assert.AreEqual(PopCommandResultCode.RequestError, deserialized.Code);
          Assert.IsNull(deserialized.StatusResponse);
          Assert.IsNull(deserialized.ResponseText);
          Assert.IsNotNull(deserialized.Description);

          var responses = deserialized.ReceivedResponses.ToArray();

          Assert.AreEqual(0, responses.Length);
        });
      });
    }
  }
}
